####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from sqlalchemy import orm
from zope import interface

from ore.contentmirror import schema, interfaces


class PeerFactory(object):

    interface.implements(interfaces.IPeerFactory)

    def __init__(self, context, transformer):
        self.context = context
        self.transformer = transformer

    @property
    def name(self):
        return self.context.__class__.__name__ + 'Peer'

    def make(self):
        klass = type(self.name, (schema.Content,),
                     dict(transformer=self.transformer))

        # With single value references creating additional foreign
        # keys to the content table, we need to distinguish the
        # join condition for the class inheritance.
        if self.transformer.table:
            # unit tests exercise a custom transformer without a table.
            join_clause = (
                self.transformer.table.c.content_id == \
                schema.content.c.content_id)
        else:
            join_clause = None

        orm.mapper(klass,
                   self.transformer.table,
                   properties=dict(self.transformer.properties),
                   inherits=schema.Content,
                   inherit_condition=join_clause,
                   polymorphic_on=schema.content.c.object_type,
                   polymorphic_identity=self.name)
        return klass


class PeerRegistry(object):

    interface.implements(interfaces.IPeerRegistry)

    def __init__(self):
        self._peer_classes = {}

    def __setitem__(self, key, value):
        self._peer_classes[key] = value

    def __getitem__(self, key):
        try:
            return self._peer_classes[key]
        except KeyError:
            for base in key.mro():
                factory = self._peer_classes.get(base, None)
                if factory is not None:
                    self._peer_classes[key] = factory # cache
                    break
            return self._peer_classes[key]

    def __contains__(self, key):
        # must not check base classes here
        return key in self._peer_classes

    def items(self):
        return self._peer_classes.items()



from Products.PloneTestCase.layer import PloneSite
from Products.Five import zcml

from sqlalchemy import create_engine

from ore.contentmirror.tests.base import reset_db, test_db_uri


class MirrorLayer(PloneSite):

    @classmethod
    def setUp(cls):
        from ore.contentmirror import ptests, schema
        zcml.load_config("testing.zcml", ptests)
        schema.metadata.bind = create_engine(test_db_uri())
        schema.metadata.create_all()

    @classmethod
    def tearDown(cls):
        reset_db()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""

cms content model

"""
import os
from zope import interface
from sqlalchemy import orm, create_engine
import sqlalchemy as rdb

from ore.contentmirror.session import Session
from ore.contentmirror import interfaces

metadata = rdb.MetaData()


def environ_db():
    if os.environ.get('CONTENTMIRROR_URI'):
        metadata.bind=create_engine(os.environ['CONTENTMIRROR_URI'])

environ_db() # set database from the environment

ContentSequence = rdb.Sequence('content_sequence')

content = rdb.Table(
   "content",
   metadata,
   rdb.Column("content_id", rdb.Integer, ContentSequence, primary_key=True),
   rdb.Column("id", rdb.String(256), nullable=False),
   rdb.Column("content_uid", rdb.String(36), nullable=False),
   rdb.Column("object_type", rdb.String(64)),
   rdb.Column("status", rdb.String(64)),
   rdb.Column("portal_type", rdb.String(64)),
   rdb.Column("folder_position", rdb.Integer),
   rdb.Column("container_id", rdb.Integer,
              rdb.ForeignKey('content.content_id', ondelete="CASCADE")),
   rdb.Column("path", rdb.Text),
   rdb.Column("relative_path", rdb.Text),
   # dublin core
   rdb.Column("title", rdb.Text),
   rdb.Column("description", rdb.Text),
   rdb.Column("subject", rdb.Text),
   rdb.Column("location", rdb.Text(4000)),
   rdb.Column("contributors", rdb.Text),
   rdb.Column("creators", rdb.Text),
   rdb.Column("creation_date", rdb.DateTime),
   rdb.Column("modification_date", rdb.DateTime),
   rdb.Column("effectivedate", rdb.DateTime),
   rdb.Column("expirationdate", rdb.DateTime),
   rdb.Column("language", rdb.Text(32)),
   rdb.Column("rights", rdb.Text),
   # plone
   rdb.Column("excludefromnav", rdb.Boolean))


rdb.Index('content_content_uid_idx', content.c.content_uid, unique=True)
rdb.Index('content_modification_date_idx', content.c.modification_date)

relations = rdb.Table(
   "relations",
   metadata,
   rdb.Column("source_id", rdb.Integer,
               rdb.ForeignKey('content.content_id', ondelete='CASCADE'),
               primary_key=True),
   rdb.Column("target_id", rdb.Integer,
               rdb.ForeignKey('content.content_id', ondelete='CASCADE'),
               primary_key=True),
   rdb.Column("relationship", rdb.String(128), primary_key=True))

files = rdb.Table(
    "files",
    metadata,
    rdb.Column("content_id", rdb.Integer,
               rdb.ForeignKey('content.content_id', ondelete="CASCADE"),
               primary_key=True),
    rdb.Column("attribute", rdb.String(156), primary_key=True),
    rdb.Column("type", rdb.String(30)),
    rdb.Column("content", rdb.Binary),
    rdb.Column("path", rdb.String(300)),
    rdb.Column("file_size", rdb.Integer),
    rdb.Column("checksum", rdb.String(33)),
    rdb.Column("file_name", rdb.String(156)),
    rdb.Column("mime_type", rdb.String(80)),
   )

rdb.Index('files_idx',
      files.c.content_id,
      files.c.attribute,
      unique=True)


class Content(object):
    type = "content"
    interface.implements(interfaces.IContentPeer)


class Relation(object):

    def __init__(self, source=None, target=None, relation=None):
        self.source = source
        self.target = target
        self.relationship = relation

orm.mapper(Content, content,
            polymorphic_on=content.c.object_type,
            polymorphic_identity='content',
            properties = {
               'children': orm.relation(
                   Content,
                   backref=orm.backref('parent',
                                       remote_side=[content.c.content_id])),
               'relations': orm.relation(
                   Relation,
                   cascade="all, delete-orphan",
                   primaryjoin=(content.c.content_id==relations.c.source_id),
                   backref=orm.backref("source",
                                       remote_side=[relations.c.source_id]))})
#               'backrefs': orm.relation(
#                   Relation,
#                   primaryjoin=(content.c.content_id==relations.c.target_id),
#                   backref=orm.backref("target",
#                                       remote_side=[relations.c.target_id]))})


#class Node( object ): pass
#orm.mapper( Node, content )

class UIDFilter(object):

    __slots__ = ('uid',)

    def __init__(self, uid):
        self.uid = uid

    def __call__(self, ob):
        if not interfaces.IContentPeer.providedBy(ob):
            return False
        return ob.content_uid == self.uid


def fromUID(content_uid):
    """
    fetch an object by uid, does not flush the session, and also checks
    against new objects in the sqlalchemy session.
    """
    session = Session()
    peers = filter(UIDFilter(content_uid), session.new)
    if peers:
        return peers.pop()
    return session.query(Content).autoflush(False).filter(
        content.c.content_uid == content_uid).first()


orm.mapper(Relation, relations,
           properties = {
#               'source': orm.relation(
#                   Content, uselist=False, backref='relations',
#                   primaryjoin=content.c.content_id==relations.c.source_id),
               'target': orm.relation(
                   Content, uselist=False,
                   primaryjoin=content.c.content_id==relations.c.target_id)})


class File(object):
    """A Database contained file. """

orm.mapper(File, files,
           polymorphic_on=files.c.type,
           polymorphic_identity='db-file')



"""
Plone functional tests for serialization of container order
"""

from unittest import defaultTestLoader

import transaction
import sqlalchemy as rdb

from OFS.interfaces import IOrderedContainer

from ore.contentmirror import schema
from ore.contentmirror.ptests.base import MirrorTestCase


class ContainerPositionTest(MirrorTestCase):

    def setUp(self):
        super(ContainerPositionTest, self).setUp()
        self.loginAsPortalOwner()
        self.portal.invokeFactory("Folder", "sample")
        self.folder = self.portal["sample"]
        self.folder.invokeFactory("Document", "doc-a")
        self.folder.invokeFactory("Document", "doc-b")
        self.assertEqual(self.folder.objectIds(), ["doc-a", "doc-b"])
        transaction.commit()

    def _query_position(self, ids=("doc-a", "doc-b"), ids_only=True):
        fields = [schema.content.c.folder_position, schema.content.c.id]
        if ids_only:
            fields.pop(0)
        query = rdb.select(fields).where(
            schema.content.c.id.in_(ids)).order_by(
            schema.content.c.folder_position)
        results = list(query.execute())
        if ids_only:
            return [r[0] for r in results]
        return results

    def test_ordered_container_interface(self):
        """Verify the OrderedContainer interface being used."""
        self.assertTrue(IOrderedContainer.providedBy(self.folder))

    def test_move_content_up(self):
        """Moving content in a container is serialized to the db."""
        self.folder.moveObjectsUp(["doc-b"])
        self.assertEqual(self.folder.objectIds(), ["doc-b", "doc-a"])
        self.folder.plone_utils.reindexOnReorder(self.folder)
        transaction.commit()
        results = self._query_position()
        self.assertEqual(results, [u'doc-b', u'doc-a'])

    def test_move_content_down(self):
        """Moving content in a container is serialized to the db."""
        self.folder.moveObjectsDown(["doc-a"])
        self.assertEqual(self.folder.objectIds(), ["doc-b", "doc-a"])
        # plone's ui does this after all moves to update the catalog
        self.folder.plone_utils.reindexOnReorder(self.folder)
        transaction.commit()
        results = self._query_position()
        self.assertEqual(results, [u'doc-b', u'doc-a'])

    def test_portal_move_content_down(self):
        """
        we have to utilize an additional operation factory and serializer
        for the portal root since its not serializable.
        """
        self.portal.invokeFactory("Document", "doc-c")
        self.assertEqual(self.portal.objectIds()[-1], "doc-c")
        transaction.commit()# serialize the content
        self.portal.moveObjectsToTop(["doc-c"])
        self.portal.plone_utils.reindexOnReorder(self.portal)
        transaction.commit()
        results = self._query_position(ids=("doc-c",), ids_only=False)
        # it moves to the top of the content in the portal
        self.assertEqual(results, [(59, u"doc-c")])
        self.assertEquals(self.portal.contentIds()[0], "doc-c")


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
The Model Loader drives the startup bootstrapping, its wired into the
zcml directives, its responsible for loading/generating the tables,
and creating the rdb mapped peer classes.

"""
from zope import component
from ore.contentmirror import schema, interfaces, plone


class ModelLoader(object):

    def __init__(self, metadata):
        self.metadata = metadata

    def load(self, klass):
        plone.push(klass) # simulate required plone environment for content
        try:
            instance = klass("transient")
        finally:
            plone.pop(klass) # remove simulated environment

        registry = component.queryUtility(interfaces.IPeerRegistry)
        if klass in registry:
            raise KeyError("Duplicate %r"%klass)

        transformer = self.transform(instance)
        peer_class = self.peer(instance, transformer)

        registry[klass] = peer_class

    def transform(self, instance):
        transformer = component.getMultiAdapter((instance, self.metadata),
                                                interfaces.ISchemaTransformer)
        transformer.transform()
        return transformer

    def peer(self, instance, transformer):
        factory = component.getMultiAdapter((instance, transformer),
                                             interfaces.IPeerFactory)
        return factory.make()

loader = ModelLoader(schema.metadata)
load = loader.load



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
######################################################################

def PortalFactory(content, operation):
    """
    filter any transient content in portal factory
    """
    # a simple check for factory in absolute url breaks portal factory..
    # no comment :-) as an alternative walk the chain.
    chain = getattr(content, 'aq_chain', None)

    if not chain:
        return

    for ob in chain[:-1]: # slice out request
        if ob.__name__ == 'portal_factory':
            operation.filtered = True
            return



"""
Functional test for content which requies access to mime types registry during
instance construction. The sample content type here is one generated by
ZopeSkel for a folder content type.
"""

from unittest import defaultTestLoader
from mocker import MockerTestCase

from zope.component import getUtility
from zope.app.component.hooks import getSite, setSite
from ore.contentmirror import loader, interfaces
from ore.contentmirror.ptests.base import MirrorTestCase

from Products.Archetypes import atapi
from Products.ATContentTypes.content import folder
from Products.ATContentTypes.content import schemata

ExampleTypeSchema = folder.ATFolderSchema.copy()
ExampleTypeSchema['title'].storage = atapi.AnnotationStorage()
ExampleTypeSchema['description'].storage = atapi.AnnotationStorage()
schemata.finalizeATCTSchema(
    ExampleTypeSchema,
    folderish=True,
    moveDiscussion=False)


class ExampleType(folder.ATFolder):
    """Description of the Example Type"""

    meta_type = "Example Type"
    schema = ExampleTypeSchema

    title = atapi.ATFieldProperty('title')
    description = atapi.ATFieldProperty('description')

    def __init__(self, *args, **kw):
        super(ExampleType, self).__init__(*args, **kw)


class ContentCreationTest(MockerTestCase, MirrorTestCase):

    def afterSetUp(self):
        """
        Remove any local sites so we recreate the environment
        in which content classes are loaded by content mirror.
        """
        local_site = getSite()
        # make sure tests are resetting environ
        self.assertTrue(local_site)
        setSite()
        self.addCleanup(setSite, local_site)

    def test_content_creation_exception(self):
        """
        Without the plone site as a local site manager attempting
        to create the content will raise an exception.
        """
        self.assertRaises(AttributeError, ExampleType, "doc-a")

    def test_content_load(self):
        loader.load(ExampleType)
        registry = getUtility(interfaces.IPeerRegistry)
        self.assertTrue(ExampleType in registry)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

import sqlalchemy as rdb
from zope import interface, component
from zope.app.container.interfaces import IContainer

try: # mock test environment compatibility
    from OFS.interfaces import IOrderedContainer
except:
    class IOrderedContainer(interface.Interface): pass

from ore.contentmirror import schema, interfaces
from ore.contentmirror.session import Session


class Serializer(object):

    interface.implements(interfaces.ISerializer)

    def __init__(self, context):
        self.context = context

    def add(self):
        registry = component.getUtility(interfaces.IPeerRegistry)
        peer = registry[self.context.__class__]()
        session = Session()
        session.add(peer)
        self._copy(peer)
        return peer

    def update(self):
        peer = schema.fromUID(self.context.UID())
        if peer is None:
            return self.add()
        self._copy(peer)
        return peer

    def delete(self):
        peer = schema.fromUID(self.context.UID())
        if peer is None:
            return
        session = Session()
        session.delete(peer)
        session.flush()

    def move(self):
        if not IContainer.providedBy(self.context):
            return self.update()
        # update contained children paths when the container moves
        peer = schema.fromUID(self.context.UID())
        contained = Session().query(schema.Content).filter(
            rdb.and_(schema.Content.path.startswith(peer.path),
                     schema.Content.content_id != peer.content_id))

        # get old and new paths to update contained
        old_containment_path = peer.path
        old_relative_path = peer.relative_path
        interfaces.ISerializer(self.context).update()
        new_containment_path = peer.path
        new_relative_path = peer.relative_path

        # for large trees this might be more efficient sans the
        # peer.
        for content in contained:
            content.path = content.path.replace(
                old_containment_path, new_containment_path)
            content.relative_path = content.relative_path.replace(
                old_relative_path, new_relative_path)
        Session().flush()

    def reposition(self):
        """
        Only fired on Ordered Containers, reserialize the folder positions of
        all contained objects. Do so without loading the contained children
        from the ZODB.
        """
        peer = schema.fromUID(self.context.UID())
        serialize_positions(self.context, peer.content_id)

    def _copy(self, peer):
        self._copyPortalAttributes(peer)
        peer.transformer.copy(self.context, peer)
        self._copyContainment(peer)

    def _copyPortalAttributes(self, peer):
        peer.portal_type = self.context.portal_type
        peer.content_uid = self.context.UID()
        peer.id = self.context.id

        peer.path = '/'.join(self.context.getPhysicalPath())
        portal_url = getattr(self.context, 'portal_url', None)
        if portal_url:
            peer.relative_path = "/".join(
                portal_url.getRelativeContentPath(self.context))
        wf_tool = getattr(self.context, 'portal_workflow', None)
        if wf_tool is None:
            return
        peer.status = wf_tool.getCatalogVariablesFor(
            self.context).get('review_state')

        container = self.context.getParentNode()
        if not IOrderedContainer.providedBy(container):
            return

        peer.folder_position = container.getObjectPosition(
            self.context.getId())

    def _copyContainment(self, peer):
        container = self.context.getParentNode()
        if container is None:
            return
        uid = getattr(container, 'UID', None)
        if uid is None:
            return
        uid = uid()
        container_peer = schema.fromUID(uid)
        if not container_peer:
            serializer = interfaces.ISerializer(container, None)
            if not serializer:
                return
            container_peer = serializer.add()
        peer.parent = container_peer


class PortalSerializer(object):

    def __init__(self, context):
        self.context = context

    def reposition(self):
        serialize_positions(self.context, None)


def serialize_positions(container, container_id):
    contained = Session().query(schema.Content).filter(
        schema.Content.container_id == container_id)
    contained_map = dict([
        (obj.id, obj) for obj in contained])
    position = 0
    for ob_id in container.objectIds():
        if not ob_id in contained_map:
            position += 1
            continue
        content = contained_map[ob_id]
        content.folder_position = position
        position += 1
    Session().flush()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

import threading
import transaction
from zope import interface, component

from ore.contentmirror.session import Session
from ore.contentmirror import interfaces


class Operation(object):

    interface.implements(interfaces.IOperation)

    filtered = False
    precedence = None

    def __init__(self, context):
        self.context = context

    @property
    def document_id(self):
        return hash(self.context)

    def process(self):
        raise NotImplementedError("subclass responsibility")

    def __cmp__(self, other):
        """Sort so content is serialized after their containers."""
        return cmp(
            self.context.getPhysicalPath(), other.context.getPhysicalPath())


class AddOperation(Operation):

    interface.implements(interfaces.IAddOperation)
    precedence = 1

    def process(self):
        interfaces.ISerializer(self.context).add()


class UpdateOperation(Operation):

    interface.implements(interfaces.IUpdateOperation)
    precedence = 0

    def process(self):
        interfaces.ISerializer(self.context).update()


class MoveOperation(Operation):

    interface.implements(interfaces.IMoveOperation)
    precedence = -1

    def process(self):
        interfaces.ISerializer(self.context).move()


class DeleteOperation(Operation):

    interface.implements(interfaces.IDeleteOperation)
    precedence = 2

    def process(self):
        interfaces.ISerializer(self.context).delete()


class RepositionOperation(Operation):

    interface.implements(interfaces.IRepositionOperation)
    # precendence might be tricky if the container moves, a reposition
    # is always done on the container. The only operation that effectively
    # cancels a reposition is a delete. All others would need to be
    # applied in parallel. also the sort comparison is wrong by default
    # since we want containers after the content has been renamed.
    precendence = -2

    def process(self):
        interfaces.ISerializer(self.context).reposition()


class BaseOperationFactory(object):

    interface.implements(interfaces.IOperationFactory)

    __slots__ = ('context',)

    def __init__(self, context):
        self.context = context

    def _store(self, op):
        # apply filters to operation
        component.subscribers((self.context, op), interfaces.IFilter)
        if op.filtered:
            return
        get_buffer().add(op)


class OperationFactory(BaseOperationFactory):

    def add(self):
        return self._store(AddOperation(self.context))

    def update(self):
        return self._store(UpdateOperation(self.context))

    def delete(self):
        return self._store(DeleteOperation(self.context))

    def move(self):
        return self._store(MoveOperation(self.context))

    def reposition(self):
        return self._store(RepositionOperation(self.context))


class PortalOperationFactory(BaseOperationFactory):
    """
    Operation Factory for the portal. Content mirror only defines one
    one operation on the portal to capture the position of its children.
    """

    def reposition(self):
        return self._store(RepositionOperation(self.context))


class BufferManager(object):

    def __init__(self, context):
        self.context = context

    def abort(self, *args):
        self.context.clear()

    def nothing(self, *args):
        pass

    def sortKey(self):
        return "content-mirror-2008"

    tpc_abort = abort
    commit = tpc_begin = tpc_vote = tpc_finish = nothing


class OperationBufferFactory(object):
    # we componentize this so we can plugin alternate implementations
    # that can do async processing
    interface.implements(interfaces.IOperationBufferFactory)

    def new(self):
        return OperationBuffer()


class OperationBuffer(object):
    """
    an operation buffer aggregates operations across a transaction
    """

    interface.implements(interfaces.IOperationBuffer)

    def __init__(self):
        self.ops = {}
        self.registered = False

    def add(self, op):
        """
        Add an operation to the buffer, aggregating with existing operations.
        """
        doc_id = op.document_id
        previous = self.ops.get(doc_id)
        if previous is not None:
            op = self._choose(previous, op)
        if op is None: # none returned when ops cancel
            del self.ops[doc_id]
            return
        self.ops[doc_id] = op
        if not self.registered:
            self._register()

    def clear(self):
        self.ops = {}
        self.registered = False
        self.manager = None

    def flush(self):
        for op in self:
            op.process()
        Session().flush()
        self.clear()

    def get(self, doc_id):
        return self.ops.get(doc_id)

    def __iter__(self):
        ops = self.ops.values()
        ops.sort()
        for op in ops:
            yield op

    def _register(self):
        self.registered = True
        transaction.get().addBeforeCommitHook(self.flush)
        transaction.get().join(BufferManager(self))

    def _choose(self, previous, new):
        """
        for a given content object, choose one operation to perform given
        two candidates. can also return no operations.
        """
        p_kind = previous.precedence
        n_kind = new.precedence

        # if we have an add and then a delete, its an effective no-op
        if (p_kind == 1 and n_kind == 2):
            return None
        if p_kind > n_kind:
            return previous
        return new

_buffer = threading.local()


def get_buffer():
    op_buffer = getattr(_buffer, 'buffer', None)
    if op_buffer is not None:
        return op_buffer
    op_buffer = component.getUtility(interfaces.IOperationBufferFactory).new()
    _buffer.buffer = op_buffer
    return op_buffer



"""
Plone Functional test for copy&paste, cut&paste, and rename.
"""

from unittest import defaultTestLoader

import transaction

from ore.contentmirror import schema
from ore.contentmirror.ptests.base import MirrorTestCase


class CopyPasteRemoveTest(MirrorTestCase):

    def _create_content(self, id):
        self.loginAsPortalOwner()
        self.portal.invokeFactory("Document", id)
        transaction.commit()
        self.assertEqual(len(list(schema.content.select().execute())), 1)
        return self.portal[id]

    def _create_tree(self, ids):
        self.loginAsPortalOwner()
        container_id = ids.pop(0)
        self.portal.invokeFactory("Folder", container_id)
        container = self.portal[container_id]
        for content_id in ids:
            container.invokeFactory("Document", content_id)
        transaction.commit()
        return container

    def _verify_path(self, path):
        query = schema.content.select().where(
            schema.content.c.relative_path == path)
        results = query.execute()
        return list(results)

    def test_cut_paste_content(self):
        document = self._create_content("content_abc")
        transaction.commit()
        cp = self.portal.manage_cutObjects([document.getId()])
        self.portal.invokeFactory("Folder", "cut_content")
        folder = self.portal["cut_content"]
        folder.manage_pasteObjects(cp)
        transaction.commit()
        # verify number of objects
        self.assertEqual(len(list(schema.content.select().execute())), 2)
        # verify path of moved content
        results = self._verify_path("cut_content/content_abc")
        self.assertEqual(len(results), 1)

    def test_cut_paste_container(self):
        container = self._create_tree(["folder_g", "doc_b"])
        transaction.commit()
        cp = self.portal.manage_cutObjects([container.getId()])
        self.portal.invokeFactory("Folder", "cut_content_2")
        folder = self.portal["cut_content_2"]
        folder.manage_pasteObjects(cp)
        transaction.commit()
        # verify number of objects
        self.assertEqual(len(list(schema.content.select().execute())), 3)
        # verify path of moved content
        results = self._verify_path("cut_content_2/folder_g/doc_b")
        self.assertEqual(len(results), 1)

    def test_copy_paste_content(self):
        document = self._create_content("content_xyz")
        cp = self.portal.manage_copyObjects([document.getId()])
        self.portal.manage_pasteObjects(cp)
        transaction.commit()
        # verify new content added
        self.assertEqual(len(list(schema.content.select().execute())), 2)
        # verify new content path
        self.assertEqual(len(self._verify_path("copy_of_content_xyz")), 1)

    def test_copy_paste_container(self):
        """
        Copy and Pasting a container in zope, does not automatically copy the
        content within it.
        """
        container = self._create_tree(["folder_b", "doc_b"])
        transaction.commit()
        cp = self.portal.manage_copyObjects([container.getId()])
        self.portal.manage_pasteObjects(cp)
        transaction.commit()
        self.assertEqual(self.portal.copy_of_folder_b.objectIds(), ["doc_b"])
        self.assertEqual(len(list(schema.content.select().execute())), 4)
        # verify new content path
        self.assertEqual(len(self._verify_path("copy_of_folder_b/doc_b")), 1)

    def test_clone_container(self):
        """
        Cloning doesn' change archetypes uid?
        """
        container = self._create_tree(["folder_c", "doc_c"])
        transaction.commit()
        self.portal.manage_clone(container, "folder_d")
        transaction.commit()
        self.assertEqual(len(list(schema.content.select().execute())), 4)
        self.assertEqual(self.portal.folder_d.objectIds(), ["doc_c"])
        # verify new content path
        self.assertEqual(len(self._verify_path("folder_d/doc_c")), 1)

    def test_rename(self):
        document = self._create_content("content_123")
        transaction.commit()
        self.portal.manage_renameObject(document.getId(), "content_111")
        transaction.commit()
        # verify number of objects
        self.assertEqual(len(list(schema.content.select().execute())), 1)
        # verify new path
        results = self._verify_path("content_111")
        self.assertEqual(len(results), 1)

    def test_rename_container(self):
        container = self._create_tree(["folder_e", "doc_e"])
        transaction.commit()
        self.portal.manage_renameObject(container.getId(), "folder_f")
        transaction.commit()
        # verify rename of contained content in db.
        self.assertEqual(self.portal.folder_f.objectIds(), ["doc_e"])
        results = self._verify_path("folder_f/doc_e")
        self.assertEqual(len(results), 1)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



"""
Plone Environment Construction for Transient Content Creation.

Some of plone products (including those form zope skel) will attempt to utilize
utilities that depend on either acquisition or a local site before they've
been attached to the object graph. In order to cope when this when loading the
classes, we can either push and pop a local component manager with the required
local utilities. Or we can push and pop the utility directly to the class,
doing the latter is signficantly simpler and avoids having depending on
mimetypes tool in the test or runtime environment. More importantly it makes
it much easier to maintain compatibility between mulitple plone versions.
"""

import sys
from zope.event import notify
from ore.contentmirror.interfaces import ContainerOrderChanged


class MimeType(object):
    mime_type = "text/plain"
    binary = 0

    def __str__(self):
        return self.mime_type


class MimeTool(object):

    mime_type = MimeType()

    def __call__(self, data, **kw):
        return unicode(data), '', self.mime_type

    def guess_encoding(self, data):
        return "utf-8"

    def lookup(self, data):
        return [self.mime_type]

mimetypes_registry = MimeTool()


def push(cls):
    setattr(cls, "mimetypes_registry", mimetypes_registry)


def pop(cls):
    delattr(cls, "mimetypes_registry")


def reindexOnReorder(self, parent):
    """
    In order to serialize content order in the container we need to modify
    plone in order to generate an event specific this to state change.

    Move events also generate a reindex, but this is effectively a
    redundant operation as the content doesn't change position. We attempt
    to filter these out here.
    """
    self._reindexOnReorder(parent)
    if sys._getframe(1).f_code.co_name != "manage_renameObject":
        notify(ContainerOrderChanged(parent))


def patch_plone():
    try:
        from Products.CMFPlone.PloneTool import PloneTool
    except ImportError:
        return

    PloneTool._reindexOnReorder = PloneTool.reindexOnReorder
    PloneTool.reindexOnReorder = reindexOnReorder

patch_plone()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

import interfaces


def objectAdded(object, event):
    operations = interfaces.IOperationFactory(object, None)
    if operations:
        operations.add()


def objectModified(object, event):
    operations = interfaces.IOperationFactory(object, None)
    if operations:
        operations.add()


def objectDeleted(object, event):
    operations = interfaces.IOperationFactory(object, None)
    if operations:
        operations.delete()


def objectMoved(object, event):
    operations = interfaces.IOperationFactory(object, None)
    if operations:
        operations.move()


def containerReordered(object, event):
    operations = interfaces.IOperationFactory(object, None)
    if operations:
        operations.reposition()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
Create the database structure for a content mirror and output to standard out.
"""

from StringIO import StringIO
from ore.contentmirror import schema
import sqlalchemy as rdb
import sys
import optparse

HELP = "usage: ./bin/instance run %prog database_type"


def main():
    parser = optparse.OptionParser(usage=HELP)
    parser.add_option('-d', '--drop', dest="drop",
                      action="store_true", default=False,
                      help="Generate Drop Tables DDL")
    parser.add_option('-n', '--nocreate', dest="create",
                      action="store_true", default=False,
                      help="Do Not Generate Create Tables DDL")

    (options, args) = parser.parse_args()

    if not len(args) == 1:
        parser.print_help()
        sys.exit(1)
        return

    db_type = args[0].strip()
    buf = StringIO()

    def write_statement(statement, parameters=''):
        ddl = statement + parameters
        buf.write(ddl.strip() + ';\n\n')

    db = rdb.create_engine('%s://'%(db_type),
                           strategy='mock',
                           executor=write_statement)
    if options.drop:
        schema.metadata.drop_all(db)
    if not options.create:
        schema.metadata.create_all(db)
    sys.stdout.write(buf.getvalue())



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from sqlalchemy import orm
import sqlalchemy as rdb
from zope import interface, component

from md5 import md5
from datetime import datetime
from ore.contentmirror.session import Session
from ore.contentmirror import interfaces, schema


class SchemaTransformer(object):
    """
    the schema transformer is stateful and in memory persistent, its
    responsible for parsing the at schema and transforming it to a
    database table.

    its also responsible for copying an object state from an instance to
    a peer.
    """
    interface.implements(interfaces.ISchemaTransformer)

    def __init__(self, context, metadata):
        self.context = context
        self.metadata = metadata
        self.table = None
        self.properties = {}

    def transform(self):
        columns = list(self.columns())
        self.table = rdb.Table(self.name,
                               self.metadata,
                               useexisting=True,
                               *columns)
        return self.table

    def copy(self, instance, peer):
        for field in self.context.Schema().fields():
            transformer = component.queryMultiAdapter((field, self))
            transformer.copy(instance, peer)

    @property
    def name(self):
        return self.context.__class__.__name__.lower().replace(' ', '_')

    def filter(self, field):
        # return false if filter, true if ok, filter dublin core
        return not field.__name__ in interfaces.DUBLIN_CORE

    def columns(self):
        yield rdb.Column("content_id",
                         rdb.Integer,
                         rdb.ForeignKey('content.content_id',
                                        ondelete="CASCADE"),
                         primary_key=True)

        for field in self.context.Schema().fields():
            if not self.filter(field):
                continue

            transformer = component.getMultiAdapter((field, self))
            result = transformer.transform()

            if isinstance(result, rdb.Column):
                yield result


class NamedFieldTransform(object):
    """
    Encapsulates database column naming.
    """

    @property
    def name(self):
        name = self.context.__name__.lower()
        return self._get_name(name)

    def _get_name(self, name):
        name = name.replace(' ', '_')
        if name in self._reserved_names:
            name = "at_" + name
        return name

    @property
    def _reserved_names(self):
        e = self.transformer.metadata.bind
        if e is None:
            return ()
        return e.dialect.preparer.reserved_words


class BaseFieldTransformer(NamedFieldTransform):

    interface.implements(interfaces.IFieldTransformer)

    column_type = rdb.String
    column_args = ()

    def __init__(self, context, transformer):
        self.context = context
        self.transformer = transformer

    def transform(self):
        args, kw = self._extractDefaults()
        column = rdb.Column(self.name,
                            self.column_type(*self.column_args),
                            *args,
                            **kw)

        self.transformer.properties[self.name] = column
        return column

    def copy(self, instance, peer):
        accessor = self.context.getAccessor(instance)
        value = accessor and accessor()
        setattr(peer, self.name, value)

    def _extractDefaults(self):
        args = []
        #if use_field_default and field.default:
        #    args.append( rdb.PassiveDefault( field.default ) )
        kwargs = {} #{
        #'nullable' : not self.context.required,
        #    'key' : self.context.getName(),
        #    }
        return args, kwargs


class ComputedTransform(BaseFieldTransformer):
    component.adapts(interfaces.IComputedField, interfaces.ISchemaTransformer)
    column_type = rdb.Text
    column_args = ()


class StringTransform(BaseFieldTransformer):
    component.adapts(interfaces.IStringField, interfaces.ISchemaTransformer)
    column_type = rdb.Text
    column_args = ()

    def copy(self, instance, peer):
        accessor = self.context.getAccessor(instance)
        value = accessor and accessor()

        # one of a number of common bad practices that plone permits is specing
        # a widget which doesn't correspond to the field. ie. using a lines
        # widget with a string field actually generates a list value for the
        # field. sadly this sort of sloppy usage appears to be rather common
        # place, so we'll try to handle it inline.. at least for now.
        if isinstance(value, (tuple, list)):
            return LinesTransform(self.context, self.transformer).copy(
                instance, peer)
        # at least morph it into a value which won't cause further issues.
        if not isinstance(value, basestring):
            value = str(value)

        setattr(peer, self.name, value)


class TextTransform(BaseFieldTransformer):
    component.adapts(interfaces.IStringField, interfaces.ISchemaTransformer)
    column_type = rdb.Text
    column_args = ()


class LinesTransform(StringTransform):
    component.adapts(interfaces.ILinesField, interfaces.ISchemaTransformer)

    def copy(self, instance, peer):
        accessor = self.context.getAccessor(instance)
        value = accessor and accessor()
        if isinstance(value, (list, tuple)):
            value = "\n".join(value)
        setattr(peer, self.name, value)


class FileTransform(NamedFieldTransform):
    """
    a file field serializer that utilizes a file peer.
    """
    interface.implements(interfaces.IFieldTransformer)
    component.adapts(interfaces.IFileField, interfaces.ISchemaTransformer)

    def __init__(self, context, transformer):
        self.context = context
        self.transformer = transformer

    def transform(self):
        file_orm = orm.relation(
            schema.File,
            cascade='all',
            uselist=False,
            primaryjoin=rdb.and_(
                schema.files.c.content_id == schema.content.c.content_id,
                schema.files.c.attribute == self.name))

        self.transformer.properties[self.name] = file_orm

    def copy(self, instance, peer):
        accessor = self.context.getAccessor(instance)
        value = accessor and accessor()
        if not value:
            return
        file_peer = self._getPeer(instance, peer)
        if not self._checkModified(value, file_peer):
            return
        self._copyPeer(file_peer, instance, value)
        setattr(peer, self.name, file_peer)

    def new(self):
        return schema.File()

    def _checkModified(self, value, peer):
        checksum = md5(str(value.data)).hexdigest()
        if peer.checksum == checksum:
            return False
        return True

    def _getPeer(self, instance, peer):
        file_peer = getattr(peer, self.name, None)
        if not file_peer:
            return self.new()
        return file_peer

    def _copyPeer(self, file_peer, instance, value):
        file_peer.attribute = self.name
        file_peer.content = str(value.data)
        file_peer.mime_type = self.context.getContentType(instance)
        file_peer.file_size = len(file_peer.content)
        file_peer.file_name = getattr(value, 'filename', value.getId())
        file_peer.checksum = md5(file_peer.content).hexdigest()


class ImageTransform(FileTransform):
    component.adapts(interfaces.IImageField, interfaces.ISchemaTransformer)


class PhotoTransform(FileTransform):
    component.adapts(interfaces.IPhotoField, interfaces.ISchemaTransformer)


class BooleanTransform(BaseFieldTransformer):
    component.adapts(interfaces.IBooleanField, interfaces.ISchemaTransformer)
    column_type = rdb.Boolean


class IntegerTransform(BaseFieldTransformer):
    component.adapts(interfaces.IIntegerField, interfaces.ISchemaTransformer)
    column_type = rdb.Integer


class FloatTransform(BaseFieldTransformer):
    component.adapts(interfaces.IFloatField, interfaces.ISchemaTransformer)
    column_type = rdb.Float


class FixedPointTransform(BaseFieldTransformer):
    component.adapts(interfaces.IFixedPointField,
                     interfaces.ISchemaTransformer)
    column_type = rdb.Float


class DateTimeTransform(BaseFieldTransformer):
    column_type = rdb.DateTime
    component.adapts(interfaces.IDateTimeField, interfaces.ISchemaTransformer)

    def copy(self, instance, peer):
        value = self.context.getAccessor(instance)()
        if not value:
            return
        value = datetime.fromtimestamp(value.timeTime())
        setattr(peer, self.name, value)


class ReferenceTransform(NamedFieldTransform):
    component.adapts(interfaces.IReferenceField, interfaces.ISchemaTransformer)
    interface.implements(interfaces.IReferenceField)

    def __init__(self, context, transformer):
        self.context = context
        self.transformer = transformer

    def transform(self):
        """
        For multi-valued references do nothing, as we'll use the generic
        relation table. For single valued references we'll directly create
        a foreign key reference to the content table.
        """
        if self.context.multiValued:
            return
        column_name = self.name+'_id'
        column = rdb.Column(
            column_name, rdb.Integer(),
            rdb.ForeignKey("content.content_id", ondelete="SET NULL"))
        relation = orm.relation(
            schema.Content,
            uselist=False,
            primaryjoin=(column == schema.content.c.content_id))
        self.transformer.properties[self.name] = relation
        self.transformer.properties[column_name] = column
        return column

    def copySingleValue(self, instance, peer):
        value = self.context.getAccessor(instance)()
        if isinstance(value, (list, tuple)):
            if len(value) >= 1:
                value = value[0]
            else:
                value = None
        if not value:
            setattr(peer, self.name+'_id', None)
        else:
            reference_peer = self._fetch_peer(value)
            if not reference_peer:
                return
            setattr(peer, self.name, reference_peer)

    def _fetch_peer(self, ob):
        peer_ob = schema.fromUID(ob.UID())
        if peer_ob is None:
            serializer = interfaces.ISerializer(ob, None)
            if serializer is None:
                return None
            peer_ob = serializer.add()
        return peer_ob

    def copy(self, instance, peer):
        single_value = not self.context.multiValued
        if single_value:
            return self.copySingleValue(instance, peer)

        value = self.context.getAccessor(instance)()
        if not value:
            return

        if not isinstance(value, (list, tuple)):
            value= [value]

        rel_map = dict([((r.relationship, r.target.content_uid), r)
                        for r in peer.relations])
        oids_seen = set() # oids of the current reference values of this field

        for ob in value:
            t_oid = ob.UID()
            oids_seen.add(t_oid)

            # skip if the object is already related
            related = (self.context.relationship, t_oid) in rel_map
            if related:
                continue

            # fetch the remote side's peer
            reference_peer = self._fetch_peer(ob)
            if not reference_peer:
                continue

            # create the relation
            relation = schema.Relation(peer,
                                       reference_peer,
                                       self.context.relationship)

        # delete old values for multi valued relations
        rel_oids = set([oid for rel_type, oid in rel_map
                        if rel_type == self.context.relationship])
        for oid in (rel_oids - oids_seen):
            Session().delete(rel_map[(self.context.relationship, oid)])



from Testing.ZopeTestCase import Sandboxed
from Products.PloneTestCase import PloneTestCase

from ore.contentmirror.ptests.layer import MirrorLayer
from ore.contentmirror.tests.base import reset_db
from ore.contentmirror.operation import get_buffer

PloneTestCase.setupPloneSite()


class MirrorTestCase(Sandboxed, PloneTestCase.PloneTestCase):

    layer = MirrorLayer

    events = ()

    def setUp(self):
        super(MirrorTestCase, self).setUp()
        reset_db()

    def flush(self):
        get_buffer().flush()



import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name="ore.contentmirror",
    version=read("version.txt").strip(),
    url="http://contentmirror.googlecode.com",
    install_requires=["setuptools",
                      "SQLAlchemy>=0.5.5,<0.6",
                      "zope.sqlalchemy",
                      "zope.component",
                      "zope.event",
                      "zope.schema",
#                      "zope.app.container",
#                      "zope.app.event",
#                      "zope.configuration"
#                      "zope.security",
#                      "zope.app.component",
#                      "ZODB3",
                      "mocker"],
    packages=find_packages(exclude=["*.tests"]),
    namespace_packages=["ore"],
    package_data = {
      "": ["*.txt", "*.zcml", "*.pt"],
    },
    zip_safe=False,
    author="Kapil Thangavelu",
    author_email="kapil.foss@gmail.com",
    description="Deploy/Rescue Content from Plone to a Relational Database",
    long_description=read("ore", "contentmirror", "readme.txt"),
    license="GPL",
    keywords="plone zope zope3",
    entry_points={
     "console_scripts": ["mirror-ddl = ore.contentmirror.ddl:main",
                         "mirror-bulk = ore.contentmirror.bulk:main"],
        }
    )



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from zope import interface, schema

from zope import component
#from zope.app.component.metaconfigure import utility
#from zope.component.security import PublicPermission

import sqlalchemy
from ore.contentmirror import interfaces
from ore.contentmirror import loader

from zope.configuration import fields
#from zope import interface

from zope.component import zcml


class IContentMirror(interface.Interface):

    content = fields.GlobalObject(
        title=u"Content Class to be Mirrored")

    serializer = fields.GlobalObject(
        title=u"Serializer For the Content",
        required=False)

    transformer = fields.GlobalObject(
        title=u"Schema Transformer",
        required=False)


def mirror(_context, content, serializer=None, transformer=None):

    if not interfaces.IMirrored.implementedBy(content):
        interface.classImplements(content, interfaces.IMirrored)

    if serializer:
        zcml.adapter(_context,
                     (serializer,),
                     interfaces.ISerializer,
                     (content,))

    if transformer:
        zcml.adapter(_context,
                     (transformer,),
                     interfaces.ISchemaTransformer,
                     (content, interfaces.IMetaData))

    _context.action(
        discriminator = (content, interfaces.IMirrored),
        callable = loader.load,
        args = (content,))


class IEngineDirective(interface.Interface):
    """ Creates A Database Engine. Database Engines are named utilities.
    """
    url = schema.URI(title = u'Database URL',
                     description = u'SQLAlchemy Database URL',
                     required = True)

    name = schema.Text(title = u'Engine Name',
                       description=u"Empty if this is the default engine",
                       required = False,
                       default = u'')

    echo = schema.Bool(title = u'Echo SQL statements',
                       description = u'Debugging Log for Database Engine',
                       required = False,
                       default=False)

    pool_recycle = schema.Int(title = u"Connection Recycle",
                              description=u"Time Given in Seconds",
                              required = False,
                              default = -1)


# keyword arguments to pass to the engine
IEngineDirective.setTaggedValue('keyword_arguments', True)


def engine(_context, url, name='', echo=False, pool_recycle=-1, **kwargs):

    engine_component = sqlalchemy.create_engine(
        url, echo=echo,
        pool_recycle=pool_recycle, **kwargs)

    zcml.utility(_context,
                 provides = interfaces.IDatabaseEngine,
                 component = engine_component,
                 name = name)


class IBindDirective(interface.Interface):
    """ Binds a MetaData to a database engine."""

    engine = schema.Text(title = u"Engine Name")

    metadata = fields.GlobalObject(
        title=u"Metadata Instance",
        description = u"Metadata Instance to be bound")


def bind(_context, engine, metadata):

    def _bind(engine_name, metadata):
        metadata.bind = component.getUtility(
            interfaces.IDatabaseEngine, engine)

    _context.action(
        discriminator = ('alchemist.bind', metadata),
        callable = _bind,
        args = (engine, metadata))



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
Export the Archetypes Schema Information for Mirrored Type as JSON

Advanced use cases for content mirror want to have a fuller
introspection of the origin archetypes schema.

A couple of concrete examples that we want to fulfill

 - Extracting Static Vocabularies ( as defined in
   Archetypes.Field.Field.Vocabulary ) for use by the front end.

 - Capturing database encoding of certain multi value fields. Ie. a
   linesfield, will serialize a list of strings to a string with
   newline separated values. A mirror front end may want to introspect
   this sort of information

 - Capturing schemata for customizing mirror front end display.
"""

import simplejson
import os
import sys

from ore.contentmirror import interfaces
from zope import component


def extract_vocabulary(vocabulary):

    # string denotes dynamic vocabulary based on instance method, skip
    if isinstance(vocabulary, str):
        return

    # process simple values directly specified as vocabulary
    if isinstance(vocabulary, (list, tuple)):
        if isinstance(vocabulary[0], (list, tuple)):
            return [(v[0], v[1]) for v in vocabulary]
        if isinstance(vocabulary[0], basestring):
            return [(v, v) for v in vocabulary]

    # process something displaylist like
    if interfaces.IDisplayList.providedBy(vocabulary):
        return vocabulary.items()


def serialize_schema(info):
    klass, peer_klass = info
    instance = klass('transient')
    schema = instance.Schema()

    fields = []
    for f in schema.fields():
        ftransform = component.getMultiAdapter((f, peer_klass.transformer))
        fd = dict(
            name = f.__name__,
            column_name = getattr(ftransform, 'name', ''),
            type = f.__class__.__name__,
            schemata = f.schemata,
            )
        if f.vocabulary:
            vocabulary = extract_vocabulary(f.vocabulary)
            if vocabulary:
                fd['vocabulary'] = vocabulary
        fields.append(fd)

    cd = dict(
        name = klass.__name__,
        table_name = peer_klass.transformer.name,
        fields = fields)
    return cd


def main():
    if len(sys.argv) == 2:
        fh = open(os.path.expanduser(
            os.path.expandvars(sys.argv[1].strip())), 'w')
    else:
        fh = sys.stdout

    registry = component.getUtility(interfaces.IPeerRegistry)
    values = map(serialize_schema, registry.items())
    fh.write(simplejson.dumps({'types': values}))
    fh.close()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
######################################################################
"""
./bin/zopectl run mirror-batch PORTAL_PATH

a content mirror bulk importer

"""
import transaction
import time
import sys
import optparse

import sqlalchemy as rdb
from ore.contentmirror import interfaces, session, schema
from DateTime import DateTime

# pyflakes
session


def expunge(ob):
    try:
        ob._p_deactivate()
    except:
        pass


def setup_parser():
    parser = optparse.OptionParser(
        usage="usage: ./bin/instance run ./bin/%prog [options] portal_path")
    parser.add_option(
        '-i', '--incremental', dest="incremental", action="store_true",
        help="serialize content modified/created since last run",
        default=False)
    parser.add_option(
        '-t', '--types', dest="types", default="",
        help="only serialize specified types (comma separated)")
    parser.add_option(
        '-p', '--path', dest="path", default="",
        help="serialize content from the specified path")
    parser.add_option(
        '-q', '--quiet', dest='verbose', action='store_false',
        help="quiet/silent output", default=True)
    parser.add_option(
        '-b', '--batch', dest='threshold', type="int",
        help="batch commit every N objects (default %s)"%(500), default=500)
    parser.add_option(
        '-d', '--db', dest='database', default="",
        help="use the specified database uri")
    return parser


def setup_query(options):
    """
    Setup a portal catalog query based on query parameters.

    Any filtering of content will still automatically pull in containers and
    referenced items of matched content regardless of whether they match
    the query to maintain consistency.
    """
    # sorting on created, gets us containers before children
    query = {'sort_on': 'created'}

    # sync by type
    if options.types:
        types = filter(None, [t.strip() for t in options.types.split(',')])
        if types:
            query['portal_type'] = {'query': types, 'operator': 'or'}

    # sync by folder
    if options.path:
        path = filter(None, [p.strip() for p in options.path.split(',')])
        if path:
            query['path'] = {'query': path, 'depth': 0, 'operator': 'or'}

    # incremental sync, based on most recent database content date.
    if options.incremental:
        last_sync_query = rdb.select(
            [rdb.func.max((schema.content.c.modification_date))])
        last_sync = last_sync_query.execute().scalar()
        if last_sync: # increment 1s
            time_tuple = list(last_sync.timetuple())
            time_tuple[4] = time_tuple[4]+1
            last_sync = DateTime(time.mktime(tuple(time_tuple)))
            query['modified'] = {'query': last_sync, 'range': 'min'}

    return query


def get_app():
    frame = sys._getframe(2)
    return frame.f_locals.get('app')


def main(app=None, instance_path=None, threshold=500):
    parser = setup_parser()
    options, args = parser.parse_args()

    if len(args) != 1:
        parser.print_help()
        sys.exit(1)
        return

    if app is None:
        app = get_app()

    if options.database:
        schema.metadata.bind = rdb.create_engine(options.database)

    instance_path = args[0]
    portal = app.unrestrictedTraverse(instance_path)

    count = 0 # how many objects have we processed

    start_time = time.time()
    batch_time = start_time # we track the time for each batch
    # sorting on created, gets us containers before children
    query = {'sort_on': 'created'}

    query = setup_query(options)
    for brain in portal.portal_catalog.unrestrictedSearchResults(**query):
        try:
            ob = brain.getObject()
            # skip broken objects
            if not ob.UID():
                continue
        except: # got to keep on moving
            continue

        serializer = interfaces.ISerializer(ob, None)
        if serializer is None:
            expunge(ob)
            continue
        count += 1

        # the object may have been processed by as a dependency
        # so we use update, which dispatches to add if no db
        # state is found.
        serializer.update()
        expunge(ob)

        if count % threshold == 0:
            transaction.commit()
            if options.verbose:
                output = ("Processed ", str(count), "in",
                          "%0.2f s"%(time.time()-batch_time))
                sys.stdout.write(" ".join(output)+"\n")
            ob._p_jar._cache.incrgc()
            batch_time = time.time()

    # commit the last batch
    transaction.commit()

    if options.verbose:
        output = ("Processed ", str(count), "in",
                  "%0.2f s"%(time.time()-batch_time))
        sys.stdout.write(" ".join(output)+"\n")

    if options.verbose:
        output = ("Finished in", str(time.time()-start_time))
        sys.stdout.write(" ".join(output)+"\n")



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
######################################################################
from zope import interface, schema
from zope.component.interfaces import IObjectEvent


class IDatabaseEngine(interface.Interface):
    """
    Configuration and access to a sqlalchemy pooled database connection.
    """


class IMirrored(interface.Interface):
    """
    Marker interface, signifying that the content should be mirrored
    to a database """


class IMetaData(interface.Interface):
    """
    Marker interface for sqlalchemy metadata, to allow for use
    in adaptation.
    """


# Fields already covered by the base peer content class, common to all content
DUBLIN_CORE= [
    "id",
    "contributors",
    "creators",
    "creation_date",
    "description",
    "effectiveDate",
    "expirationDate",
    "language",
    "location",
    "modification_date",
    "rights",
    "subject",
    "title",
    'path',
    # the following aren't dublin core but are common to atct, we include
    # them here to have them filtered by default from the peer table
    "allowDiscussion",
    "excludeFromNav",
    ]

#########################################
# Custom Events
#########################################

class IContainerOrderChangedEvent(IObjectEvent):
    """
    A container event fired when an object within a container has its position.
    Distinguished from container modified events as its more specific to
    positioning events, rather than generic to containment.
    """


class ContainerOrderChanged(object):
    interface.implements(IContainerOrderChangedEvent)

    def __init__(self, object):
        self.object = object

#########################################
# Runtime Serialization
#########################################

class IOperation(interface.Interface):

    oid = schema.ASCIILine(description=u"The identifier for the content")
    filtered = schema.Bool(
        description=u"denotes the operation should be skipped (filter set)")

    def process(connection):
        """
        process an index operation
        """


class IFilter(interface.Interface):
    """ a subscription adapter that can filter objects serialized"""


class IDeleteOperation(IOperation):
    """Operation for when content is deleted."""


class IUpdateOperation(IOperation):
    """Operation for when content is updated."""


class IAddOperation(IOperation):
    """Operation for when content is added."""


class IMoveOperation(IOperation):
    """Operation for when content is moved."""


class IRepositionOperation(IOperation):
    """Operation for when a container has its content ordering changed."""


class IOperationBuffer(interface.Interface):
    """ transactional operation buffer """

    def add(operation):
        """ add an operation to the buffer """

    def flush():
        """ flush the buffer and process/handle the operations """


class IOperationBufferFactory(interface.Interface):
    """ a factory for creating operation buffer instances, as buffers
        are accessed via utilities """


class IOperationFactory(interface.Interface):
        """ """


class ISerializer(interface.Interface):

    def add():
        """
        add the object to the database
        """

    def delete():
        """
        delete the object from the database
        """

    def update():
        """
        update the object state in the database
        """

    def move():
        """
        move the content's container
        """

    def reposition():
        """
        serialize a container's contained content positions
        """

########################################
## Schema Generation
########################################

class IModelLoader(interface.Interface):

    def load(klass):
        """ load a content class"""

    def transform():
        """ return a schema transformer """

    def peer(transformer):
        """ generate a content peer class for the content class """


class ISchemaTransformer(interface.Interface):
    """ translate an archetypes schema to a relational schema """

    metadata = interface.Attribute("metadata")

    def transform():
        """
        return a sqlalchemy database table representation
        """


class IFieldTransformer(interface.Interface):
    """ transforms an archetypes field into a sqlalchemy field """

    def transform():
        """
        returns the equivalent sqlalchemy field
        """

    def copy(instance, peer):
        """
        copies the field value from the instance to the peer.
        """

########################################
## Content Peers
########################################

class IContentPeer(interface.Interface):
    """
    A relational persisted class that has a mirror of attributes
    of a portal content class.
    """

    transformer = schema.Object(ISchemaTransformer)


class IContentFile(interface.Interface):
    """A File."""


class IPeerFactory(interface.Interface):
    """An object can make peer classes."""

    def make():
        """
        create a peer class, with a mapper, returns the mapped orm class
        """


class IPeerRegistry(interface.Interface):
    """ a registry mapping a content class to its orm peer class """

########################################
## Interface Specifications for AT Fields
########################################

class IComputedField(interface.Interface):
    """Archetypes Computed Field"""


class IStringField(interface.Interface):
    """Archetypes String Field"""


class IBooleanField(interface.Interface):
    """Archetypes Boolean Field"""


class IIntegerField(interface.Interface):
    """Archetypes Integer Field"""


class IFloatField(interface.Interface):
    """Archetypes Float Field"""


class IFixedPointField(interface.Interface):
    """Archetypes Fixed Point Field"""


class IReferenceField(interface.Interface):
    """Archetypes Reference Field """


class ILinesField(interface.Interface):
    """Archetypes Lines Field """


class IFileField(interface.Interface):
    """Archetypes Field Field """


class IImageField(interface.Interface):
    """Archetypes Image Field """


class IPhotoField(interface.Interface):
    """Archetypes Photo Field """


class ITextField(interface.Interface):
    """Archetypes Text Field """


class IDateTimeField(interface.Interface):
    """Archetypes DateTime Field """


class IDisplayList(interface.Interface):
    """Archetypes Vocabulary Display List """
